/*************************** ipsyslog.c **************************/

/* Sets of procedures that can be used to generate syslogs
   through UDP packets to localhost instead of using the UNIX domain
   /dev/log socket. */

/* Written in a hurry to solve a specific problem.
   I place this code in the public domain. It is free to everybody.

                        Marc Chatel, 1996 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#ifdef __alpha
#include <sys/syslog_pri.h>
#else
#include <syslog.h>
#endif
#include <string.h>
#include <errno.h>
#include <varargs.h>

#define LOGBUFLEN   2000
#define LOGBASELEN   100
#define MAXPROCNAME   30

int    init_done        = 0;
int    our_log_facility = 0;
int    our_sock_fd      = -1;

pid_t  ipsyslog_current_pid = 0;

int    logbuf_len = 0;
char   logbuf[LOGBUFLEN];
int    logbase_len = 0;
char   logbase[LOGBASELEN];

void ip_syslog();

/************************* startup_ip_syslog() *************************/

/* equivalent to openlog, but uses UDP traffic to syslog UDP port.
   Can only be called once within a program. Further calls are ignored. */

int startup_ip_syslog( proc_name, log_facility )
char *proc_name;
int   log_facility;
{
   int                 our_rc;
   int                 log_port = 0;
   struct sockaddr_in  our_sock, remote_sock;
   struct servent     *tmp_servent;

   if (init_done) return 1;

   our_log_facility = 0; our_sock_fd = -1;
   logbuf_len = 0;   logbuf[0] = '\0';
   logbase_len = 0;  logbase[0] = '\0';

   our_sock_fd = socket( AF_INET, SOCK_DGRAM, 0 );

   if (our_sock_fd < 0)
   {
      return 0;
   }

   our_sock.sin_family      = AF_INET;
   our_sock.sin_addr.s_addr = htonl(INADDR_ANY);
   our_sock.sin_port        = htons(0);

   our_rc = bind( our_sock_fd, (struct sockaddr *) &our_sock,
                               sizeof(struct sockaddr_in) );

   if (our_rc != 0)
   {
      close(our_sock_fd);
      our_sock_fd = -1;
      return 0;
   }

   tmp_servent = getservbyname( "syslog", "udp" );

   if (tmp_servent == NULL)
   {
      close(our_sock_fd);
      endservent();
      our_sock_fd = -1;
      return 0;
   }

   log_port = tmp_servent->s_port;
   endservent();

   remote_sock.sin_family      = AF_INET;
   remote_sock.sin_addr.s_addr = inet_addr( "127.0.0.1" );
   remote_sock.sin_port        = log_port;

   our_rc = connect( our_sock_fd, (struct sockaddr *) &remote_sock,
                                  sizeof(struct sockaddr_in) );

   if (our_rc != 0)
   {
      close(our_sock_fd);
      our_sock_fd = -1;
      return 0;
   }

   our_log_facility = log_facility;
   logbuf_len       = 0;
   logbuf[0]        = '\0';

   ipsyslog_current_pid = getpid();

   if (proc_name != NULL)
   {
      char *tmp_ptr = NULL;

      do
      {
         tmp_ptr = strchr( proc_name, '/' );
         if (tmp_ptr != NULL)
         {
            proc_name = tmp_ptr + 1;
            if (proc_name[0] == '\0')
            {
               close(our_sock_fd);
               our_sock_fd = -1;
               return 0;
            }
         }
      }
      while (tmp_ptr != NULL);

      strncpy( logbase, proc_name, MAXPROCNAME );
      logbase[MAXPROCNAME] = '\0';
      strcat( logbase, "[" );
      logbase_len = strlen( logbase );
      sprintf( logbase + logbase_len, "%d", (int) ipsyslog_current_pid );
      strcat( logbase, "]: " );
      logbase_len = strlen( logbase );
   }
   init_done = 1;
   return 1;
} /* startup_ip_syslog() */


/******************************** end_ip_syslog() ****************************/

/* Should be called when one syslog is over. Actually sends out the
   syslog in a UDP packet. Called by ip_syslog */

void end_ip_syslog()
{
   int our_rc;

   if (!init_done)       return;
   if (logbuf_len == 0)  return;

   logbuf[logbuf_len++] = '\n';
   logbuf[logbuf_len]   = '\0';
   our_rc = write( our_sock_fd, logbuf, logbuf_len + 1 );

   if (our_rc != (logbuf_len + 1))
   {
      /* fprintf( stderr, "SEND failed\n" ); */
      logbuf_len = 0;
      logbuf[0] = '\0';
      return;
   }
   logbuf_len = 0;
   logbuf[0] = '\0';
} /* end_ip_syslog() */


/*************************** begin_ip_syslog() **************************/

/* Should be called to start a new log record. Sets up things and
   sets the log's priority. */

void begin_ip_syslog( our_prio )
int our_prio;
{
   int tmp_fac;

   if (!init_done)
   {
      (void) startup_ip_syslog( "Unknown", LOG_DAEMON );
   }
   if (logbuf_len != 0)   end_ip_syslog();

   if ((our_prio < 0) || (our_prio > 7))
   {
      ip_syslog( LOG_ERR, "Next syslog had invalid priority=%d", our_prio );
      our_prio = LOG_ERR;
   }

   if (getpid() != ipsyslog_current_pid)
   {
      /* Process number has changed since the last call.
         Calling process is a child of the process who called us last.
         Update the log_base record. */

      ipsyslog_current_pid = getpid();

      if (logbase_len >= 6)
      {
         if (strcmp( logbase + logbase_len - 3, "]: " ) == 0)
         {
            char *tmp;

            tmp = logbase + logbase_len - 4;

            while ((tmp >= logbase) && ((*tmp) >= '0') && ((*tmp) <= '9'))
               tmp--;

            tmp++; *tmp = '\0';

            sprintf( tmp, "%d]: ", (int) ipsyslog_current_pid );

            logbase_len = strlen( logbase );
         }
      }
   }

   tmp_fac = our_log_facility + our_prio;

   sprintf( logbuf, "<%d>%s", tmp_fac, logbase );
   logbuf_len = strlen( logbuf );
} /* begin_ip_syslog() */


/***************************** ip_syslog_str() ****************************/

/* Add a string to the current syslog buffer */

void ip_syslog_str( our_str )
char *our_str;
{
   int   our_len;
   char  tmpbuf[10];

   if (!init_done)        begin_ip_syslog( LOG_INFO );
   if (logbuf_len == 0)   begin_ip_syslog( LOG_INFO );

   if (our_str == NULL)
   {
      strcpy( tmpbuf, "(null)" );
      our_str = &(tmpbuf[0]);
   }

   our_len = strlen( our_str );

   if (our_len == 0)  return;
   if ((logbuf_len + our_len) >= (LOGBUFLEN-1))
   {
      int able_len = (LOGBUFLEN - 2 - logbuf_len);
      int i;

      for (i = 0; i <= able_len; i++)
      {
         logbuf[ logbuf_len + i ] = our_str[ i ];
      }
      logbuf_len += able_len;
      return;
   }
   strcpy( logbuf + logbuf_len, our_str );
   logbuf_len += our_len;
} /* ip_syslog_str() */


/************************** ip_syslog_str_bounded() *************************/

/* Add a string to the current syslog buffer. If the string is longer than
   max_len, the string is truncated. */

void ip_syslog_str_bounded( our_str, max_len )
char *our_str;
int   max_len;
{
   int   our_len;
   char  tmpbuf[10];

   if (!init_done)        begin_ip_syslog( LOG_INFO );
   if (logbuf_len == 0)   begin_ip_syslog( LOG_INFO );

   if (our_str == NULL)
   {
      strcpy( tmpbuf, "(null)" );
      our_str = &(tmpbuf[0]);
   }

   our_len = strlen( our_str );

   if (our_len == 0)        return;
   if (max_len <= 0)        return;
   if (our_len > max_len)   our_len = max_len;

   if ((logbuf_len + our_len) >= (LOGBUFLEN-1))
   {
      int able_len = (LOGBUFLEN - 2 - logbuf_len);
      int i;

      for (i = 0; i <= able_len; i++)
      {
         logbuf[ logbuf_len + i ] = our_str[ i ];
      }
      logbuf[ logbuf_len + able_len + 1 ] = '\0';
      logbuf_len += able_len;
      return;
   }
   {
      char *dest, *src;
      int   tmp_len;

      tmp_len = our_len;
      dest = logbuf + logbuf_len; src = our_str;

      while ((tmp_len > 0) && ((*src) != '\0'))
      {
         tmp_len--; *(dest++) = *(src++);
      }
      *dest = '\0';
   }
   logbuf_len += our_len;
} /* ip_syslog_str_bounded() */


/*************************** ip_syslog_int() **************************/

/* Add the decimal representation of an integer to the current syslog buffer */

void ip_syslog_int( our_int )
int our_int;
{
   char  tmpbuf[30];

   sprintf( tmpbuf, "%d", our_int );
   ip_syslog_str( tmpbuf );
} /* ip_syslog_int() */


/**************************** ip_syslog_hex() **************************/

/* Add the hexadecimal representation of an integer to the current
   syslog buffer */

void ip_syslog_hex( our_int )
int our_int;
{
   char  tmpbuf[30];
   sprintf( tmpbuf, "%x", our_int );
   ip_syslog_str( tmpbuf );
} /* ip_syslog_hex() */


/***************************** ip_syslog_octal() **************************/

/* Add the octal representation of an integer to the current syslog buffer */

void ip_syslog_octal( our_int )
int our_int;
{
   char  tmpbuf[60];
   sprintf( tmpbuf, "%o", our_int );
   ip_syslog_str( tmpbuf );
} /* ip_syslog_octal() */


/***************************** ip_syslog_long() ****************************/

/* Add the decimal representation of a long integer to the current
   syslog buffer */

void ip_syslog_long( our_long )
long our_long;
{
   char  tmpbuf[50];

   sprintf( tmpbuf, "%ld", our_long );
   ip_syslog_str( tmpbuf );
} /* ip_syslog_long() */


/***************************** ip_syslog_char() *****************************/

/* Add the ASCII representation of a character to the current syslog buffer */

void ip_syslog_char( our_char )
char our_char;
{
   char tmpbuf[4];

   tmpbuf[0] = our_char; tmpbuf[1] = '\0';
   ip_syslog_str( tmpbuf );
} /* ip_syslog_char() */


/***************************** ip_syslog_errmsg() ***************************/

/* Add a string representation of the contents of "errno" to the current
   syslog buffer */

void ip_syslog_errmsg()
{
   char tmpbuf[80];

   if (errno < sys_nerr)  ip_syslog_str( sys_errlist[ errno ] );
   else
   {
      sprintf( tmpbuf, "System error %d", errno );
      ip_syslog_str( tmpbuf );
   }
} /* ip_syslog_errmsg() */


/***************************** ip_syslog() *****************************/

/* Same calling structure as 4.3BSD's syslog:

      log_priority,  format_string,  parameters

   This version only supports a SUBSET of the available format string
   options. Supported format specifiers are:

      %%    Outputs a percent sign
      %d    The matching parameter is an integer. Output it in decimal
      %x    The matching parameter is an integer. Output it in hexadecimal
      %o    The matching parameter is an integer. Output it in octal
      %ld   The matching parameter is a long integer. Output it in decimal
      %c    The matching parameter is an integer. Output it in ASCII
      %m    No matching parameter. Substitute with the ASCII interpretation 
              of the current value of errno.
      %s    The matching parameter is a (char *). Output it
      %.*s  The matching parameter is an integer, the next parameter is
              a (char *). Truncate the string if necessary to the length
              specified by the integer, than output the string.
      %.ns  Where n is a sequence of one or more digits. The matching
              parameter is a (char *). Truncate the string if necessary
              to the length specified by the digits (interpreted as a decimal
              number), than output the string.
*/

void ip_syslog( va_alist )
va_dcl
{
   va_list   ap;

   char     *fmt_str     = NULL;
   char     *cur_in_fmt  = NULL;
   char     *tmp_buf_ptr = NULL;
   char      ourbuf[LOGBUFLEN];
   int       safety_filler;
   int       our_prio = 0;

   ourbuf[0] = '\0';
   va_start(ap);

   safety_filler = 0;

   our_prio = va_arg( ap, int );
   fmt_str = va_arg(ap, char *);
   if (fmt_str == NULL)
   {
      begin_ip_syslog( LOG_ERR );
      ip_syslog_str( "Syslog error: format string is NULL" );
      end_ip_syslog();
      return;
   }

   begin_ip_syslog( our_prio );
   cur_in_fmt  = fmt_str;

   while (cur_in_fmt != NULL)
   {
      tmp_buf_ptr = strchr( cur_in_fmt, '%' );
      if (tmp_buf_ptr == NULL)
      {
         /* We have not found a percent sign in the remainder of the string.
            Check if we do have a string to output. */

         if ((*cur_in_fmt) != '\0')
         {
            /* We do have a string to output */

            ip_syslog_str( cur_in_fmt );
         }
         cur_in_fmt = NULL; /* we are finished */
      }
      else
      {
         /* We have found a percent sign in the remainder of the string. */

         if (cur_in_fmt != tmp_buf_ptr)
         {
            /* The percent sign is not at the beginning of the remainder.
               We need to output the part before the remainder. */

            int   len_to_out = tmp_buf_ptr - cur_in_fmt;
            char *tmp_buf    = NULL;
            char *tmp_fmt    = cur_in_fmt;

            if (len_to_out >= LOGBUFLEN)  len_to_out = LOGBUFLEN - 1;
            tmp_buf = &(ourbuf[0]);
   
            while ((len_to_out--) > 0)
            {
               *(tmp_buf++) = *(tmp_fmt++);
            }
            *tmp_buf = '\0';
            ip_syslog_str( ourbuf );
         }

         switch (*(tmp_buf_ptr+1))
         {
            case '%': ip_syslog_char( '%' );
                      cur_in_fmt = tmp_buf_ptr + 2;
                      break;
            case 'd': {
                         int tmp1;

                         tmp1 = va_arg( ap, int );
                         ip_syslog_int( tmp1 );
                         cur_in_fmt = tmp_buf_ptr + 2;
                      }
                      break;
            case 'x': {
                         int tmp1;

                         tmp1 = va_arg( ap, int );
                         ip_syslog_hex( tmp1 );
                         cur_in_fmt = tmp_buf_ptr + 2;
                      }
                      break;
            case 'o': {
                         int tmp1;

                         tmp1 = va_arg( ap, int );
                         ip_syslog_octal( tmp1 );
                         cur_in_fmt = tmp_buf_ptr + 2;
                      }
                      break;
            case 'l': {
                         long tmp1;

                         if ((*(tmp_buf_ptr+2)) != 'd')
                         {
                            ip_syslog_str("(Invalid syslog format specifier)");
                            cur_in_fmt = tmp_buf_ptr + 2;
                         }
                         else
                         {
                            tmp1 = va_arg( ap, long );
                            ip_syslog_long( tmp1 );
                            cur_in_fmt = tmp_buf_ptr + 3;
                         }
                      }
                      break;
            case 's': {
                         char *tmp1;

                         tmp1 = va_arg( ap, char * );
                         ip_syslog_str( tmp1 );
                         cur_in_fmt = tmp_buf_ptr + 2;
                      }
                      break;
            case '.': if ((*(tmp_buf_ptr+2)) == '*')
                      {
                         if ((*(tmp_buf_ptr+3)) != 's')
                         {
                            ip_syslog_str("(Invalid syslog format specifier)");
                            cur_in_fmt = tmp_buf_ptr + 3;
                         }
                         else
                         {
                            char *tmp1;
                            int   max_len;

                            max_len = va_arg( ap, int );
                            tmp1    = va_arg( ap, char * );
                            ip_syslog_str_bounded( tmp1, max_len );
                            cur_in_fmt = tmp_buf_ptr + 4;
                         }
                      }
                      else if (((*(tmp_buf_ptr+2)) >= '1') &&
                               ((*(tmp_buf_ptr+2)) <= '9'))
                      {
                         int max_len;
 
                         tmp_buf_ptr += 2;
                         max_len = (*(tmp_buf_ptr++)) - '0';

                         while (((*tmp_buf_ptr) >= '0') &&
                                ((*tmp_buf_ptr) <= '9'))
                         {
                            max_len = max_len * 10;
                            max_len += ((*(tmp_buf_ptr++)) - '0');
                         }

                         if ((*tmp_buf_ptr) != 's')
                         {
                            ip_syslog_str("(Invalid syslog format specifier)");
                            cur_in_fmt = tmp_buf_ptr;
                         }
                         else
                         {
                            char *tmp1;

                            tmp1    = va_arg( ap, char * );
                            ip_syslog_str_bounded( tmp1, max_len );
                            cur_in_fmt = tmp_buf_ptr + 1;
                         }
                      }
                      else
                      {
                         ip_syslog_str("(Invalid syslog format specifier)");
                         cur_in_fmt = tmp_buf_ptr + 2;
                      }
                      break;
            case 'c': {
                         char  tmp1;

                         tmp1 = va_arg( ap, char );
                         ip_syslog_char( tmp1 );
                         cur_in_fmt = tmp_buf_ptr + 2;
                      }
                      break;
            case 'm': {
                         ip_syslog_errmsg();
                         cur_in_fmt = tmp_buf_ptr + 2;
                      }
                      break;
            default:  ip_syslog_str( "(Invalid syslog format specifier)" );
                      cur_in_fmt = tmp_buf_ptr + 2;
                      break;
         }
      }
   }
   end_ip_syslog();
   va_end(ap);
} /* ip_syslog */

#ifdef 0
/******************************* main() ******************************/

int main( argc, argv )
int argc;
char **argv;
{
   int   our_prio     =  0;
   int   our_facility = 0;
   int   cur_arg      =  0;

   if (argc < 3)
   {
      fprintf( stderr,
       "%s: Needs at least 3 arguments, a facility, a priority, and a string\n",
        argv[0] );
      return 0;
   }

   our_facility = atoi( argv[1] );
   our_prio     = atoi( argv[2] );

   startup_ip_syslog( argv[0], our_facility );

   begin_ip_syslog( our_prio );
   cur_arg = 3;
   while (cur_arg < argc)
   {
      ip_syslog_str( argv[cur_arg] );
      ip_syslog_char( ' ' );
      cur_arg++;
   }
   end_ip_syslog();

   begin_ip_syslog( our_prio );
   ip_syslog_str( "Try it" );
   end_ip_syslog();
} /* main() */
#endif
